<?php

namespace Drupal\editorial_access_manager;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Config\Entity\ThirdPartySettingsInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\node\NodeGrantDatabaseStorageInterface;
use Drupal\node\NodeInterface;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;

/**
 * Service description.
 */
class EditorialAccessManager {

  /**
   * Used to get entity from route parameters.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected RouteMatchInterface $routeMatch;

  /**
   * Used to load supported entity types.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Used to write and update the editorial access.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * Module settings.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * Used to recalculate assigned node's grants.
   *
   * @var \Drupal\node\NodeGrantDatabaseStorageInterface
   */
  protected NodeGrantDatabaseStorageInterface $grantDatabaseStorage;

  /**
   * Logs error during editorial access management.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * Constructs an EditorialAccessManager object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\Database\Connection $database
   *   Database service.
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   Route match service.
   * @param \Drupal\node\NodeGrantDatabaseStorageInterface $grant_database_storage
   *   Node grants database storage service.
   * @param \Psr\Log\LoggerInterface $logger
   *   Editorial access manager logger.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entityTypeManager, Connection $database, RouteMatchInterface $route_match, NodeGrantDatabaseStorageInterface $grant_database_storage, LoggerInterface $logger) {
    $this->config = $config_factory->get('editorial_access_manager.settings');
    $this->entityTypeManager = $entityTypeManager;
    $this->database = $database;
    $this->routeMatch = $route_match;
    $this->grantDatabaseStorage = $grant_database_storage;
    $this->logger = $logger;
  }

  /**
   * Check if a specific entity type is supported by the module.
   *
   * @param string $entity_type_id
   *   Entity type ID.
   *
   * @return bool
   *   TRUE if the entity type is supported (enabled in configuration).
   */
  public function isSupported(string $entity_type_id) {
    return in_array($entity_type_id, $this->getSupportedEntityTypesList());
  }

  /**
   * Get the list of supported entities.
   *
   * @return string[]
   *   Machine names of supported entities.
   */
  public function getSupportedEntityTypesList() {
    return array_keys(array_filter($this->config->get('entity_types') ?? []));
  }

  /**
   * Get the list of supported entities.
   *
   * @return \Drupal\Core\Entity\EntityTypeInterface[]
   *   Key-valued list of entity type id and entity type definition.
   */
  public function getSupportedEntityTypes() {
    $supported_entity_types = [];
    foreach ($this->getSupportedEntityTypesList() as $entity_type_id) {
      $supported_entity_types[$entity_type_id] = $this->entityTypeManager->getDefinition($entity_type_id);
    }
    return $supported_entity_types;
  }

  /**
   * Gets the assignees of a specific entity and language.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   * @param string $langcode
   *   Language code.
   *
   * @return \Drupal\user\UserInterface[]
   *   List of users.
   */
  public function getEntityAssigneesPerLanguage(EntityInterface $entity, string $langcode)  : array {
    $query = $this->database->select('editorial_access', 'ea')
      ->fields('ea', ['uid'])
      ->condition('ea.entity_type', (string) $entity->getEntityTypeId())
      ->condition('ea.entity_id', (string) $entity->id())
      ->condition('ea.langcode', $langcode);
    $query->leftJoin('users', 'u', 'u.uid = ea.uid');
    /** @var \Drupal\Core\Database\StatementInterface $query_result */
    $query_result = $query->execute();
    return array_filter(array_map(function ($uid) {
      return $this->entityTypeManager->getStorage('user')->load($uid);
    }, array_values($query_result->fetchAllKeyed(0, 0))), function ($user) {
      return $user instanceof UserInterface;
    });
  }

  /**
   * Gets the assignees of a specific entity and language.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   *
   * @return array
   *   Key is the langcode value is the user list.
   */
  public function getEntityAssigneesGroupedByLanguage(EntityInterface $entity)  : array {
    $query = $this->database->select('editorial_access', 'ea')
      ->fields('ea', ['uid', 'langcode'])
      ->condition('ea.entity_type', (string) $entity->getEntityTypeId())
      ->condition('ea.entity_id', (string) $entity->id());
    $query->leftJoin('users', 'u', 'u.uid = ea.uid');
    /** @var \Drupal\Core\Database\StatementInterface $query_result */
    $query_result = $query->execute();
    $results = $query_result->fetchAll(\PDO::FETCH_ASSOC);
    $assignees = [];
    foreach ($results as $result) {
      $assignees[$result['langcode']][] = $result['uid'];
    }
    return $assignees;
  }

  /**
   * Clear all assignees for a specific entity and language.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   * @param string $langcode
   *   Language.
   * @param bool $clear_references
   *   Set it to TRUE to clear assigned references.
   */
  public function clearEntityAssignees(EntityInterface $entity, string $langcode, bool $clear_references = FALSE) {
    $this->database->delete('editorial_access')
      ->condition('entity_type', $entity->getEntityTypeId())
      ->condition('entity_id', (string) $entity->id())
      ->condition('langcode', $langcode)
      ->execute();

    /** @var \Drupal\Core\Database\StatementInterface $other_editorial_access_query_result */
    $other_editorial_access_query_result = $this->database->select('editorial_access', 'ea')
      ->fields('ea', ['entity_id'])
      ->condition('entity_type', $entity->getEntityTypeId())
      ->condition('entity_id', (string) $entity->id())
      ->execute();

    if (empty($other_editorial_access_query_result->fetchAll()) || $clear_references) {
      $this->database->delete('editorial_access_references')
        ->condition('parent_entity_type', $entity->getEntityTypeId())
        ->condition('parent_entity_id', (string) $entity->id())
        ->execute();
    }
  }

  /**
   * Set all assignees for a specific entity and language.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   * @param string $langcode
   *   Language.
   * @param array $assignees
   *   List of assignees.
   */
  public function setEntityAssignees(EntityInterface $entity, string $langcode, array $assignees) {
    $transaction = $this->database->startTransaction('editorial_access_manager_assignees');
    try {
      $this->clearEntityAssignees($entity, $langcode, TRUE);

      foreach ($assignees as $assignee) {
        $this->database->insert('editorial_access')
          ->fields(['entity_type', 'entity_id', 'langcode', 'uid', 'date'], [
            $entity->getEntityTypeId(),
            $entity->id(),
            $langcode,
            $assignee,
            time(),
          ])
          ->execute();
      }
      $this->storeEntityAssignableReferences($entity, (int) $assignee);
    }
    catch (\Exception $e) {
      $this->logger->error(sprintf('There was an error setting entity assignees: %s', $e->getMessage()));
      $transaction->rollBack();
    }
  }

  /**
   * Store all the assignable references.
   *
   * By storing this, access check can be granted for inherit entity types.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity to store its references.
   * @param int $assignee
   *   Assignee.
   */
  protected function storeEntityAssignableReferences(EntityInterface $entity, int $assignee) {
    $bundle_entity_type = $entity->getEntityType()->getBundleEntityType();
    $entity_bundle = !empty($bundle_entity_type) ? $this->entityTypeManager->getStorage($bundle_entity_type)->load($entity->bundle()) : NULL;
    if ($entity_bundle instanceof ConfigEntityBase) {
      $allowed_entity_references = $entity_bundle->getThirdPartySetting('editorial_access_manager', 'entity_references_enabled', []);
      $fields = $entity instanceof ContentEntityInterface ? $entity->getFields() : [];
      $entities = [];
      foreach ($fields as $field) {
        if ($field instanceof EntityReferenceFieldItemListInterface) {
          foreach ($field->referencedEntities() as $referenced_entity) {
            if (in_array($referenced_entity->getEntityTypeId(), $allowed_entity_references)) {
              if (!isset($entities[$referenced_entity->getEntityTypeId()])) {
                $entities[$referenced_entity->getEntityTypeId()] = [];
              }
              if (!in_array($referenced_entity->id(), $entities[$referenced_entity->getEntityTypeId()])) {
                $entities[$referenced_entity->getEntityTypeId()][] = $referenced_entity->id();
              }
            }
          }
        }
      }
      foreach ($entities as $reference_entity_type => $reference_entity_ids) {
        foreach ($reference_entity_ids as $reference_entity_id) {
          $this->database->insert('editorial_access_references')
            ->fields(['entity_type', 'entity_id', 'parent_entity_type', 'parent_entity_id', 'uid'], [
              $reference_entity_type,
              $reference_entity_id,
              $entity->getEntityTypeId(),
              $entity->id(),
              $assignee,
            ])
            ->execute();
        }
      }
    }
  }

  /**
   * Gets an entity from route parameters.
   *
   * This method is used to manage access plus
   * user assignment.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   Entity.
   */
  public function getEntityFromRouteParameters() : ?EntityInterface {
    $entity_type = $this->routeMatch->getParameter('entity_type_id');
    $entity_id = $this->routeMatch->getParameter('entity_id');
    if (!empty($entity_type) && !empty($entity_id) && $this->entityTypeManager->hasDefinition($entity_type)) {
      return $this->entityTypeManager->getStorage($entity_type)->load($entity_id);
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getTranslationOverviewAccess(EntityInterface $entity, AccountInterface $user) {
    $access = $this->userIsAssignable($entity, $user)
      && $this->userIsAssignee($entity, $user);
    if (!$access) {
      $access = $this->accessCheckAssigneeParents($entity, $user, function (EntityInterface $parent_entity) use ($user) {
        return $this->userIsAssignable($parent_entity, $user)
          && $this->userIsAssignee($parent_entity, $user);
      });
    }
    return $access;
  }

  /**
   * {@inheritdoc}
   */
  public function getTranslationAccess(EntityInterface $entity, AccountInterface $user, LanguageInterface $language) {
    $access = $this->userIsAssignable($entity, $user)
      && $this->userIsAssigneeAtLanguage($entity, $user, $language);

    if (!$access) {
      $access = $this->accessCheckAssigneeParents($entity, $user, function (EntityInterface $parent_entity) use ($user, $language) {
        return $this->userIsAssignable($parent_entity, $user)
          && $this->userIsAssigneeAtLanguage($parent_entity, $user, $language);
      });
    }

    return $access;
  }

  /**
   * Get access to the entity.
   *
   * Called in hook_entity_access.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity to check its access.
   * @param string $operation
   *   Operation to perform.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Account.
   * @param \Drupal\Core\Language\LanguageInterface $language
   *   Language.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Neutral or Allowed if user has access.
   */
  public function getEntityAccess(EntityInterface $entity, string $operation, AccountInterface $account, LanguageInterface $language) {
    $access = $this->checkEntityAccess($entity, $operation, $account, $language);
    if (!$access->isAllowed()) {
      $access = $this->accessCheckAssigneeParents($entity, $account, function (EntityInterface $parent_entity) use ($operation, $account, $language) {
        return $this->checkEntityAccess($parent_entity, $operation, $account, $language);
      }, TRUE);
    }
    return $access;
  }

  /**
   * Check entity access.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity to check its access.
   * @param string $operation
   *   Operation to perform.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   User.
   * @param \Drupal\Core\Language\LanguageInterface $language
   *   Language.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Neutral or allowed if it has access.
   */
  protected function checkEntityAccess(EntityInterface $entity, string $operation, AccountInterface $account, LanguageInterface $language) {
    $bundle_entity_type = (string) $entity->getEntityType()->getBundleEntityType();
    if (($account->hasPermission('edit assigned entity')
        || $account->hasPermission(sprintf('edit assigned %s', $entity->getEntityTypeId())))
      && in_array($operation, ['update', 'view']) && !empty($bundle_entity_type)) {
      /** @var \Drupal\Core\Config\Entity\ConfigEntityBase $entity_type */
      $entity_type = $this->entityTypeManager
        ->getStorage($bundle_entity_type)
        ->load($entity->bundle());

      // Ensure user has access to edit entity.
      if ($entity_type->getThirdPartySetting('editorial_access_manager', 'enabled', FALSE)
        && $entity instanceof TranslatableInterface) {

        /** @var \Drupal\Core\Database\StatementInterface $query_result_update */

        $query_result_update = $this->database->select('editorial_access', 'ea')
          ->fields('ea', ['entity_id'])
          ->condition('entity_type', $entity->getEntityTypeId())
          ->condition('entity_id', (string) $entity->id())
          ->condition('langcode', $language->getId())
          ->execute();

        if (!empty($query_result_update->fetchField()) && $operation == 'update') {
          return AccessResult::allowed();
        }

        // ALlow user to view content is assigned.
        if ($operation == 'view') {
          /** @var \Drupal\Core\Database\StatementInterface $query_result_view */
          $query_result_view = $this->database->select('editorial_access', 'ea')
            ->fields('ea', ['entity_id'])
            ->condition('entity_type', $entity->getEntityTypeId())
            ->condition('entity_id', (string) $entity->id())
            ->execute();
          if (!empty($query_result_view->fetchField())) {
            return AccessResult::allowed();
          }
        }

      }
    }

    return AccessResult::neutral();
  }

  /**
   * Check if user is assignee at any of the languages of the entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Translatable entity.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User.
   *
   * @return bool
   *   TRUE when user is assignee of the entity in at least a language.
   */
  protected function userIsAssignee(EntityInterface $entity, AccountInterface $user) {
    /** @var \Drupal\Core\Database\StatementInterface $query_result */
    $query_result = $this->database->select('editorial_access', 'ea')
      ->fields('ea', ['uid'])
      ->condition('ea.entity_type', (string) $entity->getEntityTypeId())
      ->condition('ea.entity_id', (string) $entity->id())
      ->condition('ea.uid', (string) $user->id())
      ->execute();
    return !empty($query_result->fetchField());
  }

  /**
   * Check if user is assignee at a specific entity language.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Translatable entity.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User.
   * @param \Drupal\Core\Language\LanguageInterface $language
   *   Entity language.
   *
   * @return bool
   *   TRUE when user is assignee of the entity in the specific language.
   */
  protected function userIsAssigneeAtLanguage(EntityInterface $entity, AccountInterface $user, LanguageInterface $language) {
    $assignees = $this->getEntityAssigneesPerLanguage($entity, $language->getId());
    foreach ($assignees as $assignee) {
      if ($user->id() == $assignee->id()) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Check access to the assigned parents.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity to check its access.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User.
   * @param callable $callback
   *   Callback that execute the access check.
   * @param bool $return_as_object
   *   Set it to true if it is needed
   *   to return access result object instead of a boolean.
   *
   * @return bool|AccessResultInterface
   *   TRUE/AccessResultAllowed when the user has access
   *   to perform the operation (callback) on the parent.
   */
  public function accessCheckAssigneeParents(EntityInterface $entity, AccountInterface $user, callable $callback, bool $return_as_object = FALSE) {
    $parents = $this->database->select('editorial_access_references', 'ear')
      ->fields('ear', ['parent_entity_type', 'parent_entity_id'])
      ->condition('entity_type', $entity->getEntityTypeId())
      ->condition('entity_id', $entity->id())
      ->condition('uid', $user->id())
      ->execute()
      ->fetchAll(\PDO::FETCH_ASSOC);
    foreach ($parents as $parent) {
      $parent_entity = $this->entityTypeManager->getStorage($parent['parent_entity_type'])->load($parent['parent_entity_id']);
      if ($entity instanceof EntityInterface && $this->entityTypeSupportsEntityParentEditAccess($parent_entity, $entity->getEntityTypeId())) {
        $access = $callback($parent_entity);
        if ($access instanceof AccessResultInterface && $access->isAllowed()) {
          return $access;
        }
      }
    }
    return $return_as_object ? AccessResult::neutral() : FALSE;
  }

  /**
   * Checks referenced entity type can be assigned on specific entity parent.
   *
   * Entity parent is the entity that is assigned to a user and is referencing
   * a specific entity type, which is needed to check it is supported.
   *
   * Method used to grant access to referenced entity only
   * when settings are enabled.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity to check entity type referenced is supported.
   * @param string $entity_type
   *   Entity type needed to check.
   *
   * @return bool
   *   TRUE when entity type is set up in third party settings
   *   of the entity's bundle.
   */
  protected function entityTypeSupportsEntityParentEditAccess(EntityInterface $entity, $entity_type) {
    $bundle_entity_type = $entity->getEntityType()->getBundleEntityType();
    $bundle = $this->entityTypeManager->getStorage($bundle_entity_type)->load($entity->bundle());
    return $bundle instanceof ThirdPartySettingsInterface
      && in_array($entity_type, $bundle->getThirdPartySetting('editorial_access_manager', 'entity_references_enabled', []));
  }

  /**
   * Check if the entity edit / translation can be assigned.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User.
   *
   * @return bool
   *   TRUE when the user has global or granular permissions
   *   to edit the assigned entity.
   */
  public function userIsAssignable(EntityInterface $entity, AccountInterface $user) {
    $bundle_entity_type = $entity->getEntityType()->getBundleEntityType();

    if (!empty($bundle_entity_type)) {
      $bundle_entity_type = $this->entityTypeManager->getStorage($bundle_entity_type)->load($entity->bundle());
      return ($bundle_entity_type instanceof ThirdPartySettingsInterface
          && $bundle_entity_type->getThirdPartySetting('editorial_access_manager', 'enabled', FALSE))
        && ($user->hasPermission('edit assigned entity')
          || $user->hasPermission(sprintf('edit assigned %s', $entity->getEntityTypeId())));
    }

    return FALSE;
  }

  /**
   * Recalculate grants of a specific node.
   *
   * Use this when after changing editorial access for a node.
   *
   * @param \Drupal\node\NodeInterface $node
   *   Node which grants will be recalculated.
   */
  public function recalculateNodeGrants(NodeInterface $node) {
    /** @var \Drupal\node\NodeAccessControlHandlerInterface $access_control_handler */
    $access_control_handler = $this->entityTypeManager->getAccessControlHandler('node');
    $grants = $access_control_handler->acquireGrants($node);
    $this->grantDatabaseStorage->write($node, $grants);
  }

  /**
   * Recalculate entity assignees.
   *
   * It is needed so that when an entity having entity references is updated, if
   * there are new entity references those are updated in the database.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   */
  public function recalculateEntityAssignees(EntityInterface $entity) {
    $assignees_by_language = $this->getEntityAssigneesGroupedByLanguage($entity);
    foreach ($assignees_by_language as $langcode => $assignees) {
      $this->setEntityAssignees($entity, $langcode, $assignees);
    }
  }

}
