<?php

namespace Drupal\Tests\editorial_access_manager\Functional;

use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\media\MediaInterface;
use Drupal\media\MediaTypeInterface;
use Drupal\Tests\media\Traits\MediaTypeCreationTrait;

/**
 * Test that proper permissions grants access to editorial manager access page.
 */
class EditorialAccessManagerAssignmentReferenceTest extends EditorialAccessManagerTestsBase {

  use MediaTypeCreationTrait;

  /**
   * Media type of the created media.
   *
   * @var \Drupal\media\MediaTypeInterface
   */
  protected MediaTypeInterface $mediaType;

  /**
   * Media created on test data creation.
   *
   * @var \Drupal\media\MediaInterface
   */
  protected MediaInterface $media;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'system',
    'node',
    'user',
    'entity_test',
    'language',
    'content_translation',
    'editorial_access_manager',
    'editorial_access_manager_test',
    'media',
    'media_test_source',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setupBundle() {
    parent::setupBundle();
    $this->mediaType = $this->createMediaType('test');
  }

  /**
   * {@inheritdoc}
   */
  protected function setupTestFields() {
    parent::setupTestFields();

    $field_name = 'field_media';
    $field_label = 'media';
    FieldStorageConfig::create([
      'field_name' => $field_name,
      'type' => 'entity_reference',
      'entity_type' => $this->entityTypeId,
      'cardinality' => 1,
      'translatable' => 1,
      'settings' => [
        'target_type' => 'media',
      ],
    ])->save();
    FieldConfig::create([
      'entity_type' => $this->entityTypeId,
      'field_name' => $field_name,
      'bundle' => $this->bundle,
      'label' => $field_label,
      'translatable' => 1,
      'settings' => [
        'handler' => 'default:media',
      ],
    ])->save();
  }

  /**
   * Test a entity reference of assigned entity can be edited.
   */
  public function testEditReferencedEntity() {
    $label = 'Test entity with media';
    $entity_type = $this->entityTypeId;
    $entity = $this->drupalGetEntityTestByLabel($label, $entity_type);
    // Assert assignable user didn't hae access before.
    $this->drupalLogin($this->assignableTranslator);

    $this->drupalGet($this->media->toUrl('edit-form'));
    $this->assertSession()->statusCodeEquals(403);

    \Drupal::service('editorial_access_manager.manager')
      ->setEntityAssignees($entity, $entity->language()->getId(), [$this->assignableTranslator->id()]);

    $this->drupalGet($this->media->toUrl('edit-form'));
    $this->assertSession()->statusCodeEquals(200);

    // Ensure if entity reference is disabled the user will lose the access.
    /** @var \Drupal\Core\Config\Entity\ConfigEntityBundleBase $bundle_entity_type */
    $bundle_entity_type = \Drupal::entityTypeManager()->getStorage($entity->getEntityType()->getBundleEntityType())->load($entity->bundle());
    $bundle_entity_type->setThirdPartySetting('editorial_access_manager', 'entity_references_enabled', []);
    $bundle_entity_type->save();

    $this->drupalGet($this->media->toUrl('edit-form'));
    $this->assertSession()->statusCodeEquals(403);
  }

  /**
   * {@inheritdoc}
   */
  protected function getConfiguredEntityTypes() : array {
    return [
      'node' => [
        [
          'bundle' => 'page',
        ],
      ],
      'entity_test_mul_with_bundle' => [
        [
          'bundle' => 'entity_test_mul_with_bundle',
          'entity_references_enabled' => [
            'media',
          ],
        ],
      ],
    ];
  }

  /**
   * Create entities to do the tests.
   */
  protected function createTestData() {

    /** @var \Drupal\media\MediaInterface $media */
    $media = \Drupal::entityTypeManager()
      ->getStorage('media')
      ->create(['bundle' => $this->mediaType->id()]);
    $this->media = $media;
    $this->media->save();

    $this->createEntity([
      'name' => 'Test entity with media',
      'field_eamanager_text' => $this->randomString(),
      'field_media' => $this->media->id(),
    ], 'en');

  }

}
