<?php

namespace Drupal\Tests\editorial_access_manager\Functional;

use Behat\Mink\Element\NodeElement;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Url;

/**
 * Test that proper permissions grants access to editorial manager access page.
 */
class EditorialAccessManagerAssignmentTest extends EditorialAccessManagerTestsBase {

  /**
   * Examples to test the access to page selection.
   */
  public function dataProviderTestAssignmentPageSelectionAccess() {
    return [
      'Translator manager: Assign Edition Allowed, Assign Translation Allowed' => [
        'translator_manager',
        TRUE,
        TRUE,
      ],
      'Translator manager Entity test Mul: Assign Edition Allowed, Assign Translation forbidden' => [
        'translator_entity_test_mul_manager',
        TRUE,
        FALSE,
      ],
      'Translator manager Entity test Mul Translation: Assign Edition Forbidden, Assign Translation Allowed' => [
        'translator_entity_test_mul_translation_manager',
        FALSE,
        TRUE,
      ],
      'Translator manager Node: Assign Edition Forbidden, Assign Translation forbidden' => [
        'translator_node_manager',
        FALSE,
        FALSE,
      ],
    ];
  }

  /**
   * Tests the assignment page.
   *
   * @param string $user_name
   *   User name.
   * @param bool $edit_allowed
   *   TRUE to assert edit must be allowed, FALSE otherwise.
   * @param bool $translation_allowed
   *   TRUE to assert translation must be allowed, FALSE otherwise.
   *
   * @dataProvider dataProviderTestAssignmentPageSelectionAccess
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function testAssignmentPageSelectionAccess(string $user_name, bool $edit_allowed, bool $translation_allowed) {
    $user = $this->getUserByName($user_name);
    $this->drupalLogin($user);
    $entity = $this->drupalGetEntityTestByLabel('Test entity with translation assignment access');
    $this->drupalGet($entity->toUrl('drupal:editorial-access-management'));

    $this->assertEquals($edit_allowed, $this->canAssignContentToLanguage($entity, 'en'));
    $this->assertEquals($translation_allowed, $this->canAssignContentToLanguage($entity, 'fr'));
    $this->assertEquals($translation_allowed, $this->canAssignContentToLanguage($entity, 'it'));
  }

  /**
   * Get in a editorial management page the assignment popup link.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   Entity.
   * @param string $langcode
   *   Language.
   */
  protected function getAssignmentLink(ContentEntityInterface $entity, string $langcode) {
    $url = Url::fromRoute('editorial_access_manager.editorial_assignment', [
      'entity_type_id' => $entity->getEntityTypeId(),
      'entity_id' => $entity->id(),
      'langcode' => $langcode,
    ]);
    return $this->getSession()->getPage()->find('xpath', sprintf('//table//a[starts-with(@href, "%s")]', $url->toString()));
  }

  /**
   * Check if it is possible to assign content in a specific language.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   Entity.
   * @param string $langcode
   *   Language.
   *
   * @return bool
   *   TRUE when it is possible to assign content in a specific language.
   */
  protected function canAssignContentToLanguage(ContentEntityInterface $entity, string $langcode) {
    $assignment_link = $this->getAssignmentLink($entity, $langcode);
    return $assignment_link instanceof NodeElement;
  }

  /**
   * Enables translation for the current entity type and bundle.
   */
  protected function enableTranslation() {
    parent::enableTranslation();
    \Drupal::service('content_translation.manager')->setEnabled('node', 'page', TRUE);
  }

  /**
   * Examples to test the assignment page.
   */
  public function dataProviderTestAssignmentPageSelectionAssignableUsers() {
    return [
      [
        'Test entity with translation assignment access',
        'entity_test_mul_with_bundle',
      ],
      [
        'Test node with translation assignment access',
        'node',
      ],
    ];
  }

  /**
   * Test assignment works for users.
   *
   * It is tested with a test entity and a node, because
   * node grants must be checked.
   *
   * @param string $label
   *   Entity that will be tested.
   * @param string $entity_type
   *   Entity type.
   *
   * @NOTE: consider splitting this test into several tests as several
   * flows are being checked.
   *
   * @dataProvider dataProviderTestAssignmentPageSelectionAssignableUsers
   */
  public function testAssignment(string $label, string $entity_type) {
    $entity = $this->drupalGetEntityTestByLabel($label, $entity_type);
    $french_language = \Drupal::languageManager()->getLanguage('fr');
    // Assert assignable user didn't hae access before.
    $this->drupalLogin($this->assignableTranslator);

    $this->drupalGet($entity->toUrl('drupal:content-translation-add')
      ->setRouteParameter('source', 'en')
      ->setRouteParameter('target', 'fr'));
    $this->assertSession()->statusCodeEquals(403);

    // Check editor can assign content to an assignable user.
    $this->drupalLogin($this->translatorManager);
    $this->drupalGet($entity->toUrl('drupal:editorial-access-management'));
    $assignment_link = $this->getAssignmentLink($entity, 'fr');
    $assignment_link->click();
    $assignable_user = $this->assignableTranslator;
    $page = $this->getSession()->getPage();
    $page->fillField('edit-users-0', sprintf('%s (%s)', $assignable_user->getDisplayName(), $assignable_user->id()));
    $page->pressButton('Save');

    // CHeck assignable user can create and edit translations.
    $this->drupalLogin($assignable_user);
    $this->drupalGet('/admin/content/assigned');
    $this->assertSession()->pageTextContains($label);
    $this->assertSession()->pageTextContains('Create translation');
    $this->clickLink('Create translation');

    $this->assertSession()->pageTextContains(sprintf('Create French translation of %s', $label));

    $french_label = sprintf('%s (fr)', $label);
    $page->fillField(ucfirst($entity->getEntityType()->getKey('label')), $french_label);
    $page->pressButton('Save');

    $this->drupalGet('/admin/content/assigned');

    // Check the entity translation has been correctly modified.
    $entity = $this->drupalGetEntityTestByLabel($label, $entity_type, TRUE);
    $this->assertTrue($entity->hasTranslation('fr'));
    $this->assertEquals($entity->label(), $label);
    $this->assertEquals($entity->getTranslation('fr')->label(), $french_label);

    $this->assertSession()->pageTextContains($french_label);
    $this->assertSession()->pageTextContains('Edit translation');
    $this->clickLink('Edit translation');

    $this->assertSession()->pageTextContains(sprintf('%s [French translation]', $french_label));

    // Assert that after removing access the user does not have access anymore.
    $this->drupalLogin($this->translatorManager);
    $this->drupalGet($entity->toUrl('drupal:editorial-access-management'));
    $assignment_link = $this->getAssignmentLink($entity, 'fr');
    $assignment_link->click();
    $page->fillField('edit-users-0', '');
    $page->pressButton('Save');

    $this->drupalLogin($assignable_user);
    $this->drupalGet('/admin/content/assigned');
    $this->assertSession()->pageTextNotContains($french_label);
    $this->assertSession()->pageTextContains('There is no assigned content yet.');
    $this->drupalGet($entity->toUrl('drupal:content-translation-edit', [
      'language' => $french_language,
    ]));
    $this->assertSession()->statusCodeEquals(403);
  }

  /**
   * Data set to check the access to unpublished content works.
   */
  public function dataProviderTestAccessToUnpublishedContent() {
    return [
      'Edit assigned entities: Access allowed' => [
        'assignable_translator',
        200,
      ],
      'Edit assigned nodes: Access allowed' => [
        'assignable_node',
        200,
      ],
      'Edit assigned entity test mul with bundle: Access denied' => [
        'assignable_entity_test_mul',
        403,
      ],
    ];
  }

  /**
   * Test users can access unpublished content after assignment has een done.
   *
   * @dataProvider dataProviderTestAccessToUnpublishedContent
   */
  public function testAccessToUnpublishedContent(string $user_name, int $status_code) {
    $node = $this->drupalGetEntityTestByLabel('Test node unpublished with assignment access', 'node');
    $user = $this->getUserByName($user_name);
    $this->drupalLogin($user);
    $this->drupalGet($node->toUrl());
    $this->assertSession()->statusCodeEquals(403);

    \Drupal::service('editorial_access_manager.manager')
      ->setEntityAssignees($node, 'fr', [$user->id()]);

    $this->drupalGet($node->toUrl());
    $this->assertSession()->statusCodeEquals($status_code);
  }

  /**
   * {@inheritdoc}
   */
  protected function createTestData() {
    parent::createTestData();

    $this->drupalCreateNode([
      'title' => 'Test node with translation assignment access',
      'type' => 'page',
    ]);

    $this->drupalCreateNode([
      'title' => 'Test node unpublished with assignment access',
      'type' => 'page',
      'status' => FALSE,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  protected function getConfiguredEntityTypes() : array {
    return [
      'node' => [
        [
          'bundle' => 'page',
        ],
      ],
      'entity_test_mul_with_bundle' => [
        [
          'bundle' => 'entity_test_mul_with_bundle',
        ],
      ],
    ];
  }

}
