<?php

namespace Drupal\Tests\editorial_access_manager\Functional;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\entity_test\Entity\EntityTestMulBundle;
use Drupal\node\Entity\NodeType;
use Drupal\Tests\content_translation\Functional\ContentTranslationTestBase;
use Drupal\user\UserInterface;

/**
 * Base functional tests class.
 *
 * It provides:
 *   - Testing fields
 *   - Module basic and extensible configuration.
 *   - User able to assign content.
 *   - User able to get content assigned.
 */
abstract class EditorialAccessManagerTestsBase extends ContentTranslationTestBase {

  use StringTranslationTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'system',
    'node',
    'user',
    'entity_test',
    'language',
    'content_translation',
    'editorial_access_manager',
    'editorial_access_manager_test',
  ];

  /**
   * User able to assign content.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $translatorManager;

  /**
   * User that edit any assigned entity.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $assignableTranslator;

  /**
   * User that edit any assigned entity test mul.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $assignableEntityTestMul;

  /**
   * Anonymous user that checks access denied.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $anonymousUser;

  /**
   * User able to assign entity test mul.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $translatorEntityTestMulManager;

  /**
   * User able to assign entity test mul.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $translatorEntityTestMulTranslationManager;

  /**
   * User able to assign nodes.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $translatorNodeManager;

  /**
   * User that edit any assigned node.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $assignableNode;

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected $entityTypeId = 'entity_test_mul_with_bundle';

  /**
   * {@inheritdoc}
   */
  protected $fieldName = 'field_eamanager_text';

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->doSetup();
    $this->createTestData();
  }

  /**
   * {@inheritdoc}
   */
  protected function setupBundle() {
    parent::setupBundle();
    NodeType::create([
      'type' => 'page',
      'label' => $this->t('Page'),
    ])->save();
    EntityTestMulBundle::create([
      'id' => $this->bundle,
      'label' => $this->bundle,
    ])->save();
    EntityTestMulBundle::create([
      'id' => 'other_bundle',
      'label' => $this->t('Other bundle'),
    ])->save();
  }

  /**
   * {@inheritdoc}
   */
  protected function enableTranslation() {
    parent::enableTranslation();
    $this->setupEditorialAccessManager();
    $this->rebuildAll();
  }

  /**
   * Setup module configuration.
   */
  protected function setupEditorialAccessManager() {
    $configured_entity_types = $this->getConfiguredEntityTypes();
    $this->config('editorial_access_manager.settings')
      ->set('entity_types', array_map(function () {
        return 1;
      }, $configured_entity_types))
      ->save();

    $entity_type_manager = \Drupal::entityTypeManager();
    foreach ($configured_entity_types as $entity_type => $bundles) {
      /** @var \Drupal\Core\Entity\EntityTypeInterface $entity_type */
      $entity_type = $entity_type_manager->getDefinition($entity_type);
      /** @var string $bundle_entity_type */
      $bundle_entity_type = $entity_type->getBundleEntityType();
      if (!empty($bundle_entity_type)) {
        $bundle_entity_type_storage = $entity_type_manager->getStorage($bundle_entity_type);
        foreach ($bundles as $bundle) {
          $bundle_info = $bundle_entity_type_storage->load($bundle['bundle']);
          $bundle_info->setThirdPartySetting('editorial_access_manager', 'enabled', TRUE);
          $entity_references_enabled = $bundle['entity_references_enabled'] ?? [];
          $bundle_info->setThirdPartySetting('editorial_access_manager', 'entity_references_enabled', $entity_references_enabled);

          $bundle_info->save();
        }
      }
    }
  }

  /**
   * List of entity types configured to be synchronizable.
   *
   * @return array
   *   By default, the test entity with the test fields.
   */
  protected function getConfiguredEntityTypes() : array {
    return [
      'entity_test_mul_with_bundle' => [],
      'node' => [],
    ];
  }

  /**
   * Creates and activates translator, editor and admin users.
   */
  protected function setupUsers() {
    $this->translatorManager = $this->drupalCreateUser(
      array_merge(
        $this->getEntityTestViewPermissions(), [
          'assign entity edition',
          'assign entity translation',
          'reassign assigned entities',
        ]),
      'translator_manager',
    );

    $this->assignableTranslator = $this->drupalCreateUser(
      [
        'edit assigned entity',
      ],
      'assignable_translator',
    );

    $this->assignableNode = $this->drupalCreateUser(
      [
        'edit assigned node',
      ],
      'assignable_node',
    );
    $this->assignableEntityTestMul = $this->drupalCreateUser(
      [
        'edit assigned entity_test_mul_with_bundle',
      ],
      'assignable_entity_test_mul_with_bundle',
    );

    $this->anonymousUser = $this->drupalCreateUser(
      [],
      'anonymous',
    );

    $this->translatorEntityTestMulManager = $this->drupalCreateUser(
      array_merge(
        $this->getEntityTestViewPermissions(), [
          'assign entity_test_mul_with_bundle edition',
        ]),
      'translator_entity_test_mul_manager',
    );

    $this->translatorEntityTestMulTranslationManager = $this->drupalCreateUser(
      array_merge(
        $this->getEntityTestViewPermissions(), [
          'assign entity_test_mul_with_bundle translation',
        ]),
      'translator_entity_test_mul_translation_manager',
    );

    $this->translatorNodeManager = $this->drupalCreateUser(
      array_merge(
        $this->getEntityTestViewPermissions(), [
          'assign node translation',
        ]),
      'translator_node_manager',
    );
  }

  /**
   * Get permissions to view the test entity.
   *
   * It is needed as being able to synchronize the entity implies
   * being able to view it.
   *
   * @return array
   *   Permissions to view entity + translations.
   */
  protected function getEntityTestViewPermissions() {
    return [
      'view test entity',
      'view test entity translations',
    ];
  }

  /**
   * Get an entity from the database based on its label.
   *
   * @param string|\Drupal\Component\Render\MarkupInterface $name
   *   An entity label, usually generated by $this->randomMachineName().
   * @param string $entity_type_id
   *   Entity type ID.
   * @param bool $reset
   *   (optional) Whether to reset the entity cache.
   *
   * @throw \InvalidArgumentException
   *   When the entity does not exist.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   An entity matching $title.
   */
  protected function drupalGetEntityTestByLabel($name, $entity_type_id = NULL, $reset = FALSE) {
    if (empty($entity_type_id)) {
      $entity_type_id = $this->entityTypeId;
    }
    if ($reset) {
      \Drupal::entityTypeManager()->getStorage($entity_type_id)->resetCache();
    }
    // Cast MarkupInterface objects to string.
    $entity_type = \Drupal::entityTypeManager()->getDefinition($entity_type_id);
    $name = (string) $name;
    $entities = \Drupal::entityTypeManager()
      ->getStorage($entity_type_id)
      ->loadByProperties([$entity_type->getKey('label') => $name]);
    // Load the first node returned from the database.
    $returned_entity = reset($entities);
    if (!$returned_entity instanceof EntityInterface) {
      throw new \InvalidArgumentException(sprintf('Entity with label "%s" does not exist.', $name));
    }
    return $returned_entity;
  }

  /**
   * Create entities to do the tests.
   */
  protected function createTestData() {

    $this->createEntity([
      'name' => 'Test entity with translation assignment access',
      'field_eamanager_text' => $this->randomString(),
    ], 'en');

    $this->createEntity([
      'name' => 'Test entity without translation assignment access',
      'field_eamanager_text' => $this->randomString(),
    ], 'en', 'other_bundle');

    $this->drupalCreateNode([
      'title' => 'Test node without translation assignment access',
      'type' => 'page',
    ]);
  }

  /**
   * Get one of the users created during setup by its name.
   *
   * @param string $name
   *   User name.
   *
   * @return \Drupal\user\UserInterface
   *   User.
   */
  protected function getUserByName(string $name) {

    switch ($name) {
      case 'anonymous':
        return $this->anonymousUser;

      case 'translator_manager':
        return $this->translatorManager;

      case 'translator_entity_test_mul_manager':
        return $this->translatorEntityTestMulManager;

      case 'translator_entity_test_mul_translation_manager':
        return $this->translatorEntityTestMulTranslationManager;

      case 'translator_node_manager':
        return $this->translatorNodeManager;

      case 'assignable_translator':
        return $this->assignableTranslator;

      case 'assignable_node':
        return $this->assignableNode;

      case 'assignable_entity_test_mul':
        return $this->assignableEntityTestMul;

      default:
        throw new \InvalidArgumentException(sprintf('User with name "%s" does not exists."', $name));
    }
  }

}
